"""
AixWeather core_data_format_2_output_file Package

The 'core_data_format_2_output_file' package within AixWeather contains modules for transforming
the core_data to the desired output format and actually returning the output file.
The Input here is always in the core_data format.
Each module is designed to produce a specific output format.
"""